/*********************************************************************
*                    SEGGER Microcontroller GmbH                     *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*            (c) 2014 - 2019 SEGGER Microcontroller GmbH             *
*                                                                    *
*           www.segger.com     Support: support@segger.com           *
*                                                                    *
**********************************************************************
*                                                                    *
* All rights reserved.                                               *
*                                                                    *
* Redistribution and use in source and binary forms, with or         *
* without modification, are permitted provided that the following    *
* conditions are met:                                                *
*                                                                    *
* - Redistributions of source code must retain the above copyright   *
*   notice, this list of conditions and the following disclaimer.    *
*                                                                    *
* - Neither the name of SEGGER Microcontroller GmbH                  *
*   nor the names of its contributors may be used to endorse or      *
*   promote products derived from this software without specific     *
*   prior written permission.                                        *
*                                                                    *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND             *
* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,        *
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF           *
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE           *
* DISCLAIMED.                                                        *
* IN NO EVENT SHALL SEGGER Microcontroller GmbH BE LIABLE FOR        *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR           *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT  *
* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;    *
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF      *
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT          *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE  *
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH   *
* DAMAGE.                                                            *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

File    : main.c
Purpose : Low level semihosting example

*/
#include <string.h>

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

#define SYS_OPEN        0x01
#define SYS_CLOSE       0x02
#define SYS_WRITEC      0x03
#define SYS_WRITE0      0x04
#define SYS_WRITE       0x05
#define SYS_READ        0x06
#define SYS_READC       0x07
#define SYS_ISTTY       0x09
#define SYS_SEEK        0x0A
#define SYS_FLEN        0x0C
#define SYS_GET_CMDLINE 0x15

#define angel_SWIreason_ReportException 0x18

#define ADP_Stopped_ApplicationExit 	((2 << 16) + 38)
#define ADP_Stopped_RunTimeError 	((2 << 16) + 35)

                                          // fopen-equivalent   Explanation
#define SYS_FILE_FLAG_READ            0   // "r"                Open the file for reading
#define SYS_FILE_FLAG_READBYTE        1   // "rb"               Open the file for reading
#define SYS_FILE_FLAG_READWRITE       2   // "r+"               Open the file for reading and writing
#define SYS_FILE_FLAG_READWRITEBYTE   3   // "r+"               Open the file for reading and writing
#define SYS_FILE_FLAG_WRITE           4   // "w"                Open and truncate or create the file for writing
#define SYS_FILE_FLAG_WRITEBYTE       5   // "wb"               Open and truncate or create the file for writing
#define SYS_FILE_FLAG_WRITEREAD       6   // "w+"               Open and truncate or create the file for writing and reading
#define SYS_FILE_FLAG_WRITEREADBYTE   7   // "w+b"              Open and truncate or create the file for writing and reading
#define SYS_FILE_FLAG_APPEND          8   // "a"                Open or create the file for writing
#define SYS_FILE_FLAG_APPENDBYTE      9   // "ab"               Open or create the file for writing
#define SYS_FILE_FLAG_APPENDREAD      10  // "a+"               Open or create the file for writing and reading
#define SYS_FILE_FLAG_APPENDREADBYTE  11  // "a+b"              Open or create the file for writing and reading

#define FILE_HANDLE_STDIN    0
#define FILE_HANDLE_STDOUT   1
#define FILE_HANDLE_CONSOLE  0xFFFFFF00

#ifndef __NO_SYSTEM_INIT
 void SystemInit(void) {}
#endif

/*********************************************************************
*
*       BKPT
*/
int BKPT(int op, void* p1, void* p2) {
  register int r0 asm("r0");
  register int r1 asm("r1") __attribute__((unused));
  register int r2 asm("r2") __attribute__((unused));

  r0 = op;
  r1 = (int) p1;
  r2 = (int) p2;

  asm volatile(
      " bkpt 0xAB \n"
      : "=r"(r0) // out
      :// in
      :// clobber
  );
  return r0;
}

/*********************************************************************
*
*       _SH_Open
*
*  Function description
*    Opens a file on the host system.
*/
static int _SH_Open(char* pFilename, int Mode, int Len) {
  int FileHandle;
  void* block[3];

  block[0] = (void*)pFilename;
  block[1] = (void*)Mode;
  block[2] = (void*)Len;
  FileHandle = BKPT(SYS_OPEN, (void*) block, (void*) 0);

  return FileHandle;
}

/*********************************************************************
*       _SH_Close
*
*  Function description
*    Closes a file on the host system, previously opened by _SH_Open().
*/
static int _SH_Close(int FileHandle) {
  int r;
  void* block[1];

  block[0] = (void*)FileHandle;
  r = BKPT(SYS_CLOSE, (void*)block, (void*) 0);

  return r;
}

/*********************************************************************
*       _SH_Read
*
*  Function description
*    Reads a given number of bytes from a specific file on the host system.
*    File handle can also be an interactive file like "console" or "stdin".
*/
static int _SH_Read(int FileHandle, char* pBuffer, int NumBytesToRead) {
  int NumBytesLeft;
  void* block[3];

  block[0] = (void*)FileHandle;
  block[1] = (void*)pBuffer;
  block[2] = (void*)NumBytesToRead;
  NumBytesLeft = BKPT(SYS_READ, (void*) block, (void*) 0);

  return NumBytesLeft;
}

/*********************************************************************
*
*       _SH_ReadC
*
*  Function description
*    Reads a single character from the debug channel (usually stdin).
*/
static char _SH_ReadC(void) {
  char c;

  c = BKPT(SYS_READC, (void*) 0, (void*) 0);
  return c;
}

/*********************************************************************
*
*       _SH_Write
*
*  Function description
*    Writes a given number of bytes to a file on the host system, previously opened by _SH_Open().
*/
static int _SH_Write(int FileHandle, const char* pBuffer, int NumBytesToWrite) {
  int NumBytesLeft;
  void* block[3];

  block[0] = (void*)FileHandle;
  block[1] = (void*)pBuffer;
  block[2] = (void*)NumBytesToWrite;
  NumBytesLeft = BKPT(SYS_WRITE, (void*) block, (void*) 0);

  return NumBytesLeft;
}

/*********************************************************************
*
*       _SH_Seek
*
*  Function description
*    Puts the filepointer at a specific position inside a file on the host system, previously opened by _SH_Open().
*/
static int _SH_Seek(int FileHandle, int Position) {
  int r;
  void* block[2];

  block[0] = (void*)FileHandle;
  block[1] = (void*)Position;
  r = BKPT(SYS_SEEK, (void*) block, (void*) 0);

  return r;
}

/*********************************************************************
*
*       _SH_FLen
*
*  Function description
*    Gets the size of a file on the host system, previously opened by _SH_Open().
*/
static int _SH_FLen(int FileHandle) {
  int r;
  void* block[1];

  block[0] = (void*)FileHandle;
  r = BKPT(SYS_FLEN, (void*) block, (void*) 0);
  return r;
}

/*********************************************************************
*
*       _SH_WriteC
*
*  Function description
*    Writes a single character to the debug channel (usually stdout) on the host system.
*/
static void _SH_WriteC(char* ptr) {
  BKPT(SYS_WRITEC, (void*) ptr, (void*) 0);
}

/*********************************************************************
*
*       _SH_Write0
*
*  Function description
*    Writes a zero-terminated string to the debug channel (usually stdout) on the host system.
*/
static void _SH_Write0(const char* ptr) {
  BKPT(SYS_WRITE0, (void*) ptr, (void*) 0);
}

/*********************************************************************
*
*       _SH_GetCmdLine
*/
static int _SH_GetCmdLine(char* cmdline, unsigned int len) {
  int   r;
  void* block[2];

  block[0] = (void*)cmdline;
  block[1] = (void*)len;
  r = BKPT(SYS_GET_CMDLINE, (void*)block, (void*)0);
  return r;
}

/*********************************************************************
*
*       _SH_Exit
*/
static int _SH_Exit(int Signal) {
  void* block[1];

  block[0] = (void*)Signal;
  BKPT(angel_SWIreason_ReportException, (void*)block, (void*) 0);
  return 0;
}

/*********************************************************************
*
*       _GetStrLen
*/
static int _GetStrLen(const char* s) {
  int i;

  i = 0;
  if (s) {
    do {
      if (*s == 0) {
        break;
      }
      i++;
      s++;
    } while (s);
  }
  return i;
}

/*********************************************************************
*
*       main
*/
void main(int argc, char* argv[]) {
  int hFile;
  char* pMessage;
  char* pFilename;
  int r;
  hFile = 0;
  pFilename = "c:\\temp\\test_arm_semihosting.txt";  // Set Filename
  pMessage = "Hello Segger!";  // Set message
  _SH_Write0("Semihosting tests:\r\n");
  _SH_Write0("\r\n=============================\r\n");
  //
  // SYS_OPEN
  //
  _SH_Write0("\r\nSYS_OPEN (file): ");
  hFile = _SH_Open(pFilename, SYS_FILE_FLAG_WRITEREADBYTE, strlen(pFilename));
  if (hFile == -1) {
    _SH_Write0("Failed");
    goto Done;
  } else {
    _SH_Write0("OK");
  }
  _SH_Write0("\r\n=============================\r\n");
  if (hFile == -1) {
    goto Done;
  }
  //
  // SYS_Write
  //
  _SH_Write0("\r\n=============================\r\n");
  _SH_Write0("Writing complete file");
  _SH_Write0("\r\nSYS_WRITE: ");
  if (_SH_Write(hFile, pMessage, strlen(pMessage)) != 0) {
    _SH_Write0("Failed");
    goto Done;
  } else {
    _SH_Write0("OK");
  }
  _SH_Write0("\r\n=============================\r\n");
  //
  // SYS_CLOSE
  //
  _SH_Write0("\r\n=============================\r\n");
  _SH_Write0("SYS_CLOSE: ");
  r = _SH_Close(hFile);
  if (r != 0) {
    _SH_Write0("Failed");
    goto Done;
  } else {
    _SH_Write0("OK");
  }
  _SH_Write0("\r\n=============================\r\n");
  _SH_Write0("Semihosting tests done.\r\n");
Done:
  while (1);
}


/*************************** End of file ****************************/
